package ActiveState::I18N::UTF8;

use strict;
use warnings;

our $VERSION = '1.000';

use XSLoader;
XSLoader::load __PACKAGE__, $VERSION;

use Encode qw(encode decode decode_utf8 _utf8_off);
use POSIX;
use MIME::Base64 qw(decode_base64 encode_base64);
use MIME::QuotedPrint qw(decode_qp encode_qp);
use ActiveState::UTF8 qw(is_valid_utf8);

use base qw(Exporter);
our @EXPORT_OK = qw(charset_match encoding_guess
          	    utf8_encode convert_from_utf8 utf8_header 
        encode_utf8_header utf8_length);
        
my $charset;
{
    my $cs = get_charset();

    $charset = {
	ascii => {
	    id => undef,    # valid utf-8
	    charset_name => 'US-ASCII',
	    re => qr/\basc[-_]?ii\b/,
	},
	utf8 => {
	    id => undef,    # valid utf-8
	    charset_name => 'UTF-8',
	    re => qr/\butf[-_]?8\b/,
	},
	euckr => {
	    id => $cs->{euckr},
	    charset_name => 'EUC-KR',
	    re => qr/\beuc[-_]?kr\b/,
	},
	iso2022jp => {
	    id => $cs->{iso2022jp},
	    charset_name => 'ISO-2022-JP',
	    re => qr/\biso[-_]?2022[-_]?jp\b/,
            iconv_from => 1,
	},
	iso2022kr => {
	    id => $cs->{iso2022kr},
	    charset_name => 'ISO-2022-KR',
	    re => qr/\biso[-_]?2022[-_]?kr\b/,
	},
	latin1 => {
	    id => $cs->{latin1},
	    charset_name => 'ISO-8859-1',
	    re => qr/\biso[-_]?8859[-_]1\b|\blatin[-_]?1\b/,
	},
	latin2 => {
	    id => $cs->{latin2},
	    charset_name => 'ISO-8859-2',
	    re => qr/\b(iso[-_]?)?8859[-_]2\b|\blatin[-_]?2\b/,
	},
	big5hkscs => {
	    id => 'big5hkscs',
	    charset_name => 'BIG5-HKSCS',
	    re => qr/\bbig[-_]?5[-_]?hkscs\b/,
	},
	euc_jp => {
	    id => 'EUC-JP',
	    charset_name => 'EUC-JP',
	    re => qr/\beuc[-_]?jp\b/,
            iconv_from => 1,
	},
	iso_8859_8 => {
	    id => 'ISO_8859-8',
	    charset_name => 'ISO-8859-8',
	    re => qr/\biso[-_]?8859[-_]8\b/,
            iconv_from => 1,
	},
	iso_8859_11 => { # TIS-620 +
	    id => 'ISO_8859-11',
	    charset_name => 'ISO-8859-11',
	    re => qr/\biso[-_]?8859[-_]11\b|\btis[-_]?620\b/,
	},
	iso_8859_15 => {
	    id => 'ISO_8859-15',
	    charset_name => 'ISO-8859-15',
	    re => qr/\biso[-_]?8859[-_]15\b/,
            iconv_from => 1,
	},
	koi8_r => {
	    id => 'KOI8-R',
	    charset_name => 'KOI8-R',
	    re => qr/\bkoi8[-_]?r\b/,
	},
	win1250 => {
	    id => 'windows-1250',
	    charset_name => 'cp1250',
	    re => qr/\bwin(?:dows)?[-_]?1250\b|\bcp[-_]?1250\b/,
	},
	win1251 => {
	    id => 'windows-1251',
	    charset_name => 'cp1251',
	    re => qr/\bwin(?:dows)?[-_]?1251\b|\bcp[-_]?1251\b/,
	},
	win1252 => {
	    id => 'windows-1252',
	    charset_name => 'cp1252',
	    re => qr/\bwin(?:dows)?[-_]?1252\b|\bcp[-_]?1252\b/,
	},
	win1256 => {
	    id => 'windows-1256',
	    charset_name => 'cp1256',
	    re => qr/\bwin(?:dows)?[-_]?1256\b|\bcp[-_]?1256\b/,
	},
        utf16le => {
            id => 'UTF_16LE',
            charset_name => 'UTF-16LE',
            re => qr/\butf[-_]?16le\b/,
        },
        utf16be => {
            id => 'UTF_16BE',
            charset_name => 'UTF-16BE',
            re => qr/\butf[-_]?16be\b/,
        },

	# XXX See Encode::Supported/Microsoft-related naming issues
	ksc5601 => {
	    # id => $cs->{ksc5601},
	    # charset_name => 'KSC5601-RAW',
	    id => 'ksc5601',
	    charset_name => 'cp949', # microsoft extension to EUC-KR
	    re => qr/\bks[-_]?c[-_]?5601\b/,
	},
	big5 => {
	    id => 'big5',
	    charset_name => 'cp950', # microsoft extension to Big5
	    re => qr/\bbig[-_]?5(?:[^-\w]|$)/,
	},
	gb2312 => {
	    id => 'gb2312',
	    charset_name => 'cp936', # microsoft extension to GB2312
	    re => qr/\bgb[-_]?2312\b/,
	},
	shift_jis => {
	    id => 'SHIFT-JIS',
	    charset_name => 'cp932', # microsoft extension to Shift_JIS
	    re => qr/\bshift[-_]?jis\b/,
            iconv_from => 1,
	},
    };
    
    # The default iconv ids are ok in most cases, but some platforms
    # (ie. SunOS 5.6) use different ids.  This structure contains custom iconv
    # ids for these non-standard platforms.
    my @iconv_ids = (
        { 
            re  => qr/^SunOS\s+5\.6$/,
            ids => {
                shift_jis   => 'SJIS',
                euc_jp      => 'eucJP',
                iso_8859_8  => '8859-8',
                iso_8859_15 => '8859-15',
            },
            charset_names => {
                shift_jis   => 'SJIS',
                euc_jp      => 'eucJP',
                iso_8859_8  => '8859-8',
                iso_8859_15 => '8859-15',
                iso2022jp   => 'iso2022'
            },
        },
    );
    
    # Change ids of iconv charsets if they require customization.
    # For example, SunOS 5.2 uses different ids.  This section of code
    # uses the above data structure.
    my ($sysname, undef, $release, undef, undef) = POSIX::uname();    
    for my $custom_id (@iconv_ids) {
        if ("$sysname $release" =~ $custom_id->{re}) {
            for my $cs_id (keys %{$custom_id->{ids}}) {
                $charset->{$cs_id}{id} = $custom_id->{ids}{$cs_id};
            }
            for my $cs_id (keys %{$custom_id->{charset_names}}) {
                $charset->{$cs_id}{charset_name} = $custom_id->{charset_names}{$cs_id};
            }
            last;
        }
    }
}

sub charset_match {
    my $cs = lc(shift);
    return unless defined $cs;

    for (values %$charset) {	   
        return $_ if $cs =~ $_->{re};
    }
}

my $config = {
    enabled => 1, size => 256, threshold => 0.95, banner => '',

    big5 => 1, big5hkscs => 1,
    euckr => 1, ksc5601 => 1, gb2312 => 1,
    iso2022jp => 1,
};
my $guess = {
    ascii => '[\x00-\x7f]',
    big5 => '[\xa1-\xfe][\x40-\x7e\xa1-\xfe]',
    big5ascii => '[\x88-\xfe][\x40-\x7e]',
    big5hkscs => '[\x88-\xfe][\x40-\x7e\xa1-\xfe]',
    euc => '[\xa1-\xfe][\xa1-\xfe]',
    iso2022jp => '(\x1b\(B|\x1b\(J|\x1b\$@|\x1b\$B)',
};
sub encoding_guess {
    $config = $_[1]->{encoding}->{guess} if ref $_[1]->{encoding}->{guess};

    return unless $config->{enabled};
    return unless defined $_[0];

    my $src = $_[0];
    my $tmp = substr($src, 0, $config->{size});
    my $cnt = length $tmp;
    return unless $cnt;

    my ($threshold, $odds) = ($config->{threshold}, undef);

    ## ISO-2022-JP
    if($config->{iso2022jp}) {
	return $charset->{iso2022jp} if $tmp =~ /$guess->{iso2022jp}/o;
    }

    ## ASC-II
    my $ascii_cnt = 0;
    $ascii_cnt ++ while $tmp =~ /$guess->{ascii}/go;

    ## BIG5, BIG5-HKSCS, EUC-KR, & GB2312
    my ($euc_cnt, $big5_cnt, $big5ascii_cnt, $big5hkscs_cnt) = (0, 0, 0, 0);
    $euc_cnt++ while $tmp =~ /$guess->{euc}/go;
    $big5_cnt++ while $tmp =~ /$guess->{big5}/go;
    $big5ascii_cnt++ while $tmp =~ /$guess->{big5ascii}/go;
    $big5hkscs_cnt++ while $tmp =~ /$guess->{big5hkscs}/go;

    if($big5_cnt == $big5hkscs_cnt) {
	if($euc_cnt == $big5_cnt) {
	    $odds = ($ascii_cnt + $euc_cnt * 2) / $cnt;
	    if($odds > $threshold) {
		return $charset->{gb2312} if $config->{gb2312};
		return $charset->{ksc5601} if $config->{ksc5601};
		return $charset->{euckr} if $config->{euckr};
	    }
	}
	else {
	    $odds = ($ascii_cnt - $big5ascii_cnt + $big5_cnt * 2) / $cnt;
	    if($odds > $threshold) {
		return $charset->{big5} if $config->{big5};
	    }
	}
    }
    else {
	$odds = ($ascii_cnt - $big5ascii_cnt + $big5hkscs_cnt * 2) / $cnt;
	if($odds > $threshold) {
	    return $charset->{big5hkscs} if $config->{big5hkscs};
	}
    }

    ## UTF-8
    # always check for utf8.
    return $charset->{utf8} if is_valid_utf8($src);
    
    # XXX return default charset set in etc/utf8.conf or else return undef
    return $charset->{$config->{default}} if defined $config->{default};
    return undef;
}

sub utf8_encode {
    my ($h_charset, $config) = ($_[1], $_[2]);

    # $_[0] is text input to be encoded (in place)
    return unless defined $_[0];

    # If Perl already thinks this string is UTF-8, encode it into a stream of
    # UTF-8 octets and return immediately. This is guaranteed to work by the
    # Encode API, but is quite rare in practise -- it requires getting a
    # string into UTF-8 in the first place: either by using chr(>255) or by
    # using something like XML::Parser.
    if (Encode::is_utf8($_[0])) {
	$_[0] = encode("utf8", $_[0]);
	return;
    }
    my $cs;
    
    if (ref($h_charset) && $h_charset->{id}) {
	$cs = $h_charset;
    }
    elsif ($h_charset) {
	# Use what Engine says. 
	# Note: engine puts a "nul" in the end of charset string
	$h_charset =~ s/(?:\W*)$//;
	$cs = charset_match($h_charset) || {charset_name => $h_charset};
    }
    else {
        # charset is undefined or empty
        $cs = undef;
    }

    # Use Perl-5.8's cool new Encode function to translate from $cs->{name}
    # into UTF-8. Since the incoming data may be truncated, converting it
    # from the foreign charset can fail, so we eval{} that. Once we've got
    # it in Perl's internal format, we don't need to protect encoding it
    # into UTF-8, since that's guaranteed to work.
    my $intermediate = '';
    my $len = 0;
    ENCODE:
    for my $_cs ($cs, encoding_guess($_[0], $config)) {
        next unless $_cs;

        $len = length $_[0];
        while ($len > 0) {
            eval {
                $intermediate = decode($_cs->{charset_name},
                                       substr($_[0], 0, $len),
                                       Encode::FB_DEFAULT);
            };
            if ($@) {
                # truncate a byte from $_[0] and try again
                --$len;
                next;
            }
            last ENCODE;
        }
    }

    if ($len) {
        $_[0] = encode('utf8', $intermediate);
    }
    else {
	# The last resort, we give up de/en-coding and try to return a valid 
	# utf8 thing anyway.
	# Ensure valid utf8, trimming extra bytes off the end if necessary to
	# ensure we have only whole characters.
	chop($_[0]) until is_valid_utf8($_[0]);
    }
    
    return;
}

sub convert_from_utf8 {
    my ($data, $charset, $config) = @_;
    
    # iconv must be enabled for this function    
    return unless $config->{encoding}{options}{iconv_enabled};
    
    # ensure we have data to convert
    return unless defined $data;
    
    # get the charset
    my $cs = charset_match($charset);
    return unless (ref $cs);       # charset not found
    return $data unless $cs->{id}; # doesn't need conversion (utf8 or ascii)
    return unless $cs->{iconv_from} and $cs->{charset_name};    # not an iconv charset
    
    eval {
        require Text::Iconv;
        my $converter = Text::Iconv->new('UTF-8', $cs->{charset_name});
        $data = $converter->convert($data);
    };
    
    if ($@) {
        warn $cs->{charset_name} . " is unsupported by iconv";
        return;
    }
    
    return $data;
}

sub utf8_header {
    return unless defined $_[0];

    my ($cs, $et, $dc);
    my @words = split(/(=\?.*?\?.*?\?.*?\?=)/, $_[0]);

    for (@words) {
	next unless (s/[\r\n\t]+//g, $_);
	$cs = undef;
	if(/=\?(.*)\?(.*)\?(.*)\?=/s) {
            
	    ($cs, $et, $_) = ($1, lc($2), $3);
            local $^W;
	    $dc = \&decode_base64 if $et eq 'b';
	    (s/_/ /g, $dc = \&decode_qp) if $et eq 'q';

	    $_ = $dc->($_) if ref $dc;
	}
	utf8_encode($_, $cs, $_[1]);
	unless (is_valid_utf8($_)) {
	    # conversion to utf8 failed.  This can mean a number of things such
	    # as iconv not being enabled in utf8.conf or an unsupported charset.
	    # We can try to use Perl 5.8's Encode function to see if that works better.
	     
	    $_ = decode_utf8($_, 0);

	    # change substitution code point 0xFFFD into question marks for display
	    $_ =~ s/\xFF\xFD/?/g;
	    _utf8_off($_);

	    unless(is_valid_utf8($_)) {
		$_ = '[PMX Conversion Error]';
		$cs ||= "unknown charset";
		warn "failed to convert $cs to UTF-8; " .
		 "please ensure iconv is enabled in utf8.conf";
	    }
	}
    }
    # any space between =?...?= chunks should collapse
    for (my $i = 2; $i < @words; $i += 2) {
        $words[$i] =~ s/^\s+\z//;
    }
    
    $_[0] = join('', grep {defined $_ } @words);
    
    ## rfc2231 encoded attribute support ##
    
    return unless $_[0] =~ /\*=/;
    
    my $name;
    @words = split(/(;)/, $_[0]);
    for (@words) {
        if (/(.+)\*=(.*)'(.*)'(.*)/s) {
            ($name, $cs, $_) = ($1, $2, $4);
            $_ =~ s/%([\da-fA-F]{2})/pack("C", hex($1))/ge;
            utf8_encode($_, $cs, $_[1]);
            $_ = "$name=$_";
        }
    }
    $_[0] = join('', @words);
}

sub encode_utf8_header {
    my $val = shift;
    my $enc = shift || 'B';
    
    return unless is_valid_utf8($val);

    if ($enc eq 'B') {
        $val = encode_base64($val, '');
    }
    elsif ($enc eq 'Q') {
        $val = encode_qp($val, '');
    }
    else {
        die "encoding must be base64 (B) or quoted printable (Q)";
    }        

    return "=?UTF-8?$enc?$val?=";
}

sub utf8_length {
    return unless defined $_[0];

    my ($tl, $length) = (length $_[0], $_[1]);
    return $tl if (!$length) || ($length < 0) || ($tl <= $length);

    my ($tmp, $byte) = (substr($_[0], 0, $length), undef);

    while($byte = chop($tmp)) {
	## non 10xxxxxx is the boundary
	last if $length-- <= 0;
	last if (ord($byte) & 0xc0) != 0x80;
    }

    $length;
}

1;

__END__

=head1 NAME

ActiveState::I18N::UTF8 - Global subroutines for UTF-8 encoding.

=head1 SYNOPSIS

 use ActiveState::I18N::UTF8 qw(charset_match encoding_guess
	utf8_encode utf8_header utf8_length);

 $charset = charset_match($charset_string);
 $charset = encoding_guess($text, \%utf8_config);
 utf8_encode($text, $charset_string, \%utf8_config);
 utf8_header($header, \%utf8_config);
 $length = utf8_length($text, $length);

=head1 DESCRIPTION

It provides a group of subroutines to encode text to UTF-8. Supported character
sets include ASCII, ISO-2022-JP (Japanese), ISO-8859-1 (Latin-1), BIG5 and
BIG5-HKSCS (Traditional Chinese), EUC-KR and KSC5601 (Korean), and GB2312
(Simplified Chinese) and new character sets can be added through a uniform
interface.

=over

=item $charset = charset_match($charset_string);

It returns a hash reference including the canonical name for supported character
sets as key or I<undef> by default. It makes the character-set alias passed by
'$charset_string' to be mapped into the supported character set.

=item $charset = encoding_guess($text, \%utf8_config);

The hash '%utf8_config' has several items to control the encoding guess:

    'enabled': to enable or disable encoding guess
    'size': to set the length of text for encoding guess
    'threshold': to set the threshold for encoding guess
    'big5' 'big5hkscs' 'euckr' 'ksc5601' 'gb2312' 'iso2022jp':
        to control the encoding guess for each

When '%uft8_config' is not provided or I<undef>, it uses the default one.

When the character set is not assigned, it tests the first 'size' bytes of the
'$text' and determines the character set based on the odds of encoding. The
accuracy closes to 100% for supported character sets. It returns 'undef' for
unsupported character sets.

=item utf8_encode($text, $charset_string, \%utf8_config);

It calls charset_match() to get the canonical name of the character set, and
then uses the name to call the relevant encoder to encode the '$text'. When
'%utf8_config' is not provided or I<undef>, it uses the default one.

=item utf8_header($header, \%utf8_config);

It takes a raw MIME header and returns UTF-8 encoded text. It supports multiple
encoded-words and the mixing of plain text and encoded-words no matter what the
character set and encoding are.

=item encode_utf8_header($header_value, $encoding);

Encode a raw utf-8 encoded string into a format that can be used in a MIME 
header.  Basically this is the reverse of the utf8_header function.  'encoding'
can be 'B' for base64 encoding (default), or 'Q' for quoted printable.

=item $length = utf8_length($text, $length);

For the given '$text' and '$length', it returns the new length, which is at the
right boundary of multi-byte UTF-8 characters.

=back
